'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var diffMatchPatch = require('diff-match-patch');
var htmlMatcher = require('@emmetio/html-matcher');
var scanner = require('@emmetio/scanner');

const defaultOptions = {
    compact: true,
    normalizeSpace: true,
    allTokens: true,
    baseStart: 0,
    replaceThreshold: 0,
    preserveTags: [],
    inlineElements: [
        'a', 'abbr', 'acronym', 'applet', 'b', 'basefont', 'bdo',
        'big', 'br', 'button', 'cite', 'code', 'del', 'dfn', 'em', 'font', 'i',
        'iframe', 'img', 'input', 'ins', 'kbd', 'label', 'map', 'object', 'q',
        's', 'samp', 'select', 'small', 'span', 'strike', 'strong', 'sub', 'sup',
        'textarea', 'tt', 'u', 'var'
    ]
};
function createOptions(opt) {
    return htmlMatcher.createOptions(Object.assign(Object.assign({}, defaultOptions), opt));
}

const wordDelimiters = '.,…/\\!?:;()[]{}<>"\'«»“”‘’-–—\n\r\t \u00a0'.split('').map(ch => ch.charCodeAt(0));
const numberPunct = '.,'.split('').map(ch => ch.charCodeAt(0));
/**
 * Returns updated list of diff ops so that adjacent delete and insert ops include
 * word bounds
 */
function updateWordBounds(diffs) {
    diffs = [...diffs];
    const result = [];
    let i = 0;
    let start = null;
    while (i < diffs.length) {
        const cur = diffs[i];
        const next = diffs[i + 1];
        if (cur[0] === diffMatchPatch.DIFF_DELETE && next && next[0] === diffMatchPatch.DIFF_INSERT) {
            const del = [...cur];
            const ins = [...next];
            if (start) {
                // Move start word bound, if required
                moveStartBound(start, result, del, ins);
            }
            // Advance next until we find non-updated chunk with valid word bound
            i += 2;
            while (i < diffs.length) {
                const peek = diffs[i];
                if (peek[0] === diffMatchPatch.DIFF_INSERT) {
                    ins[1] += peek[1];
                }
                else if (peek[0] === diffMatchPatch.DIFF_DELETE) {
                    del[1] += peek[1];
                }
                else {
                    const bound = findRightWordBound(peek, del, ins);
                    if (bound === -1) {
                        del[1] += peek[1];
                        ins[1] += peek[1];
                    }
                    else {
                        del[1] += peek[1].slice(0, bound);
                        ins[1] += peek[1].slice(0, bound);
                        diffs[i] = [peek[0], peek[1].slice(bound)];
                        break;
                    }
                }
                i++;
            }
            result.push(del, ins);
            start = null;
        }
        else if (cur[0] === diffMatchPatch.DIFF_DELETE && !startsWithDelimiter(cur[1]) && start) {
            const nextStart = moveStartBound(start, result, cur);
            if (start !== nextStart) {
                result.push(cur);
                result.push([diffMatchPatch.DIFF_INSERT, start[1].slice(nextStart[1].length)]);
            }
            else {
                result.push(cur);
            }
            start = null;
            i++;
        }
        else if (cur[0] === diffMatchPatch.DIFF_INSERT && !startsWithDelimiter(cur[1]) && start) {
            const nextStart = moveStartBound(start, result, cur);
            if (start !== nextStart) {
                result.push([diffMatchPatch.DIFF_DELETE, start[1].slice(nextStart[1].length)]);
                result.push(cur);
            }
            else {
                result.push(cur);
            }
            start = null;
            i++;
        }
        else {
            start = cur[0] === diffMatchPatch.DIFF_EQUAL ? cur : null;
            result.push(cur);
            i++;
        }
    }
    return result;
}
function moveStartBound(start, result, ...tokens) {
    const nextStart = updateStartBound(start, ...tokens);
    if (nextStart !== start) {
        result.pop();
        if (nextStart[1] !== '') {
            // Edge case: bound update resulted in empty prefix,
            // do not add it
            result.push(nextStart);
        }
        return nextStart;
    }
    return start;
}
/**
 * Check if character at given location in string is a word delimiter
 */
function isWordDelimiter(ch) {
    return wordDelimiters.includes(ch);
}
/**
 * Finds location of word bound in given string. Returns `-1` if word bound wasn’t
 * found
 */
function findRightWordBound(eq, del, ins) {
    const text = eq[1];
    for (let i = 0; i < text.length; i++) {
        const ch = text.charCodeAt(i);
        if (numberPunct.includes(ch)) {
            // Possibly a decimal delimiter or number formatter
            if (scanner.isNumber(text.charCodeAt(i + 1)) || scanner.isNumber(text.charCodeAt(i - 1))) {
                continue;
            }
            if (i === 0 && (scanner.isNumber(lastCharCode(del[1])) || scanner.isNumber(lastCharCode(ins[1])))) {
                continue;
            }
        }
        if (isWordDelimiter(text.charCodeAt(i))) {
            return i;
        }
    }
    return -1;
}
/**
 * Updates given diffs so they point at the beginning of the word bound
 */
function updateStartBound(start, ...tokens) {
    const end = start[1].length - 1;
    let i = end;
    while (i >= 0) {
        const ch = start[1].charCodeAt(i);
        if (numberPunct.includes(ch) && scanner.isNumber(start[1].charCodeAt(i - 1))) {
            // A decimal delimiter or formatter character of number
            i -= 2;
            continue;
        }
        if (isWordDelimiter(ch)) {
            break;
        }
        i--;
    }
    if (i !== end) {
        const offset = end - i;
        tokens.forEach(t => t[1] = start[1].slice(-offset) + t[1]);
        return [start[0], start[1].slice(0, -offset)];
    }
    return start;
}
/**
 * Check if given text starts with word delimiter
 */
function startsWithDelimiter(text) {
    return isWordDelimiter(text.charCodeAt(0));
}
function lastCharCode(text) {
    return text.charCodeAt(text.length - 1);
}

/**
 * Check if given string is whitespace-only
 */
function isWhitespace(str) {
    return str ? /^[\s\n\r]+$/.test(str) : false;
}
/**
 * Check if given token has specified `type`
 */
function isType(token, type) {
    return token ? token.type === type : false;
}
/**
 * Check if given token contains tag
 */
function isTagToken(token) {
    return token.type === 1 /* Open */
        || token.type === 2 /* Close */
        || token.type === 3 /* SelfClose */;
}
/**
 * Collects element stack for given text location
 */
function getElementStack(tokens, pos) {
    const stack = [];
    let start = 0;
    while (start < tokens.length) {
        const token = tokens[start];
        if (token.location > pos) {
            break;
        }
        if (token.location === pos && token.type === 100 /* Space */ && token.offset) {
            break;
        }
        if (token.type === 1 /* Open */) {
            stack.push(token);
        }
        else if (token.type === 2 /* Close */) {
            stack.pop();
        }
        start++;
    }
    return { stack, start };
}
/**
 * Returns last item in given array
 */
function last(arr) {
    return arr.length ? arr[arr.length - 1] : void 0;
}
/**
 * Diff token factory
 */
function diffToken(name, value, location, text = '') {
    return { type: 101 /* Diff */, name, value, location, text };
}
function isDiffToken(token) {
    return token.type === 101 /* Diff */;
}

class SliceResult {
    constructor(tokens, range) {
        this.tokens = tokens;
        this.range = range;
    }
    /**
     * Returns string representation of current fragment.
     * @param opTag Tag name for wrap markers
     */
    toString(opTag) {
        return this.tokens.map(t => {
            if (typeof t === 'string') {
                return t;
            }
            if (isSliceMark(t)) {
                return opTag ? `<${t.close ? '/' : ''}${opTag}>` : '';
            }
            return t.value;
        }).join('');
    }
    /**
     * Returns list of tag tokens
     */
    toTokens(opTag) {
        const result = [];
        for (const t of this.tokens) {
            if (typeof t !== 'string') {
                if (isSliceMark(t)) {
                    result.push({
                        type: 101 /* Diff */,
                        location: t.location,
                        name: opTag,
                        value: t.close ? `</${opTag}>` : `<${opTag}>`
                    });
                }
                else {
                    result.push(t);
                }
            }
        }
        return result;
    }
    /**
     * Converts current slice result to diff token
     */
    toDiffToken(name, text, location) {
        return diffToken(name, this.toString(name), location, text);
    }
}
/**
 * Returns document fragment for given _text content_ range that can be used as
 * a _replacement_ for the same range of given document.
 * Inserts `SliceOp` operation markers in appropriate positions to maintain valid
 * XML state.
 * @param start Token location hint in parsed document where to start fragment lookup
 */
function slice(doc, from, to, start) {
    const stack = [];
    const tokens = [sliceMark(from)];
    const range = getSliceRange(doc, from, to, start);
    const pushText = (value) => value && tokens.push(value);
    let offset = from;
    doc.tokens.slice(range[0], range[1]).forEach(token => {
        pushText(doc.content.slice(offset, token.location));
        offset = token.location;
        if (token.type === 1 /* Open */) {
            stack.push([token, tokens.length]);
            tokens.push(token);
        }
        else if (token.type === 2 /* Close */) {
            if (!stack.pop()) {
                // Closing element outside of stack
                tokens.push(sliceMark(offset, true), token, sliceMark(offset));
            }
            else {
                tokens.push(token);
            }
        }
        else {
            tokens.push(token);
        }
    });
    pushText(doc.content.slice(offset, to));
    if (isSliceMarkOpen(last(tokens))) {
        tokens.pop();
    }
    else {
        tokens.push(sliceMark(to, true));
    }
    while (stack.length) {
        // We have unclosed tags: we should add intermediate open/close markers
        const item = stack.pop();
        const pos = item[0].location;
        tokens.splice(item[1], 1, sliceMark(pos, true), item[0], sliceMark(pos));
    }
    return new SliceResult(optimize(tokens), range);
}
function fragment(doc, from, to, options = {}) {
    const { stack, start } = getElementStack(doc.tokens, from);
    let offset = from;
    let i = start;
    let end = start;
    let result = '';
    const push = (chunk) => {
        if (chunk) {
            if (typeof chunk === 'string') {
                result += chunk;
            }
            else if (isAllowedToken(chunk, options)) {
                result += chunk.value;
            }
        }
    };
    stack.forEach(push);
    // Walk up to the end of text fragment and check inner structure
    while (i < doc.tokens.length) {
        const token = doc.tokens[i];
        if (token.location > to) {
            break;
        }
        end = i++;
        push(doc.content.slice(offset, token.location));
        offset = token.location + (token.offset || 0);
        if (token.type === 1 /* Open */) {
            stack.push(token);
        }
        else if (token.type === 2 /* Close */) {
            stack.pop();
        }
        push(token);
    }
    push(doc.content.slice(offset, to));
    // Close the remaining elements
    while (stack.length) {
        const token = stack.pop();
        if (isAllowedToken(token, options)) {
            push(`</${token.name}>`);
        }
    }
    return new SliceResult([sliceMark(from), result, sliceMark(to, true)], [start, end]);
}
/**
 * A slice mark factory
 */
function sliceMark(location, close = false) {
    return {
        type: 'slice',
        location,
        close
    };
}
function isSliceMark(value) {
    return value && value.type === 'slice';
}
function isSliceMarkOpen(value) {
    return isSliceMark(value) && !value.close;
}
function isSliceMarkClose(value) {
    return isSliceMark(value) && value.close;
}
/**
 * Returns optimized token list without redundant tokens
 */
function optimize(tokens) {
    const result = [];
    for (let i = 0; i < tokens.length; i++) {
        if (isSliceMarkOpen(tokens[i]) && isSliceMarkClose(tokens[i + 1])) {
            i++;
        }
        else {
            result.push(tokens[i]);
        }
    }
    return result;
}
/**
 * Check if given token is allowed in extracted fragment according to
 * options specified
 */
function isAllowedToken(token, options) {
    if (token.type === 1 /* Open */ || token.type === 2 /* Close */ || token.type === 3 /* SelfClose */) {
        return options.tags ? options.tags.includes(token.name) : true;
    }
    return true;
}
function findTokenStart(model, pos) {
    let i = 0;
    const il = model.tokens.length;
    while (i < il) {
        const token = model.tokens[i];
        if (token.location > pos) {
            break;
        }
        // Edge case: skip closing tags from lookup precisely at given location
        if (token.location === pos && token.type !== 3 /* SelfClose */ && token.type !== 2 /* Close */) {
            break;
        }
        i++;
    }
    return i;
}
/**
 * Returns optimal range of tokens for slicing
 */
function getSliceRange(model, from, to, start = findTokenStart(model, from)) {
    const tokens = [];
    const stack = [];
    let i = start;
    while (i < model.tokens.length) {
        const token = model.tokens[i++];
        if (token.location > to) {
            break;
        }
        if (token.location === to) {
            // Do not include open tag on the edge of range
            if (token.type === 1 /* Open */) {
                break;
            }
            // Do not include not opened closing tag on the edge of range
            if (token.type === 2 /* Close */ && last(stack) !== token.name) {
                break;
            }
        }
        tokens.push(token);
        if (token.type === 1 /* Open */) {
            stack.push(token.name);
        }
        else if (token.type === 2 /* Close */) {
            stack.pop();
        }
    }
    // Remove trailing tokens
    while (tokens.length) {
        const token = tokens[0];
        // Remove unclosed tags from the beginning of range
        if (token.location === from && token.type === 1 /* Open */ && stack[0] === token.name) {
            stack.shift();
            tokens.shift();
            start++;
        }
        else {
            break;
        }
    }
    while (tokens.length) {
        const token = last(tokens);
        if (token.location && (token.type === 3 /* SelfClose */ || token.type === 100 /* Space */)) {
            tokens.pop();
        }
        else {
            break;
        }
    }
    return [start, start + tokens.length];
}

/**
 * Calculates diff between given parsed document and produces new model with diff
 * tokens in it. This model can be restored into a final XML document
 */
function diff(from, to, options = createOptions()) {
    return options.invert
        ? diffInverted(from, to, options)
        : diffNatural(from, to, options);
}
/**
 * Performs “natural” diff between given documents: compares `from` and `to` documents
 * and returns `to` document with patched XML
 */
function diffNatural(from, to, options) {
    const diffs = getDiff(from, to, options);
    const state = createState(to.tokens);
    const fragmentOpt = {
        tags: options.preserveTags || []
    };
    let toOffset = 0;
    let fromOffset = 0;
    diffs.forEach(d => {
        let value = d[1];
        if (d[0] === diffMatchPatch.DIFF_DELETE && value) {
            // Removed fragment: just add deleted content to result
            if (suppressWhitespace(value, toOffset, state)) {
                fromOffset += 1;
                value = value.slice(1);
            }
            if (!shouldSkipDel(to.content, value, fromOffset, options)) {
                // Edge case: put delete patch right after open tag or non-suppressed
                // whitespace at the same location
                moveTokensUntilPos(state, toOffset);
                const chunk = fragment(from, fromOffset, fromOffset + value.length, fragmentOpt);
                state.push(chunk.toDiffToken('del', value, toOffset));
            }
            fromOffset += value.length;
        }
        else if (d[0] === diffMatchPatch.DIFF_INSERT) {
            // Inserted fragment: should insert open and close tags at proper
            // positions and maintain valid XML nesting
            if (suppressWhitespace(value, toOffset, state)) {
                toOffset += 1;
                value = value.slice(1);
            }
            if (shouldSkipIns(value, toOffset, state, options)) {
                state.push(whitespace(toOffset, value));
            }
            else {
                moveSlice(to, toOffset, toOffset + value.length, 'ins', state);
            }
            toOffset += value.length;
        }
        else if (d[0] === diffMatchPatch.DIFF_EQUAL) {
            // Unmodified content
            toOffset += value.length;
            fromOffset += value.length;
            // Move all tokens of destination document to output result
            while (state.hasNext()) {
                const first = state.current();
                // if (first.location > offset || (first.location === offset && first.type === ElementType.Open)) {
                //     break;
                // }
                if (first.location > toOffset) {
                    break;
                }
                if (first.location === toOffset && first.type === 1 /* Open */) {
                    // Handle edge case. In the following examples:
                    // – aa <div>bb cc</div>
                    // – aa bb <div>cc</div>
                    // ...removing `bb ` results DELETE token with the same _text_
                    // location, yet in first case it should be outside `<div>` and
                    // in second case – inside `<div>`.
                    // In case if token touches the edge of open tag, we should detect
                    // if this token is inside or outside the same tag in `from`
                    // document
                    const { stack } = getElementStack(from.tokens, fromOffset);
                    let found = false;
                    while (stack.length) {
                        if (stack.pop().name === first.name) {
                            found = true;
                            break;
                        }
                    }
                    if (!found) {
                        break;
                    }
                }
                state.push(first);
                state.ptr++;
            }
        }
    });
    return {
        tokens: state.output.concat(state.input.slice(state.ptr)),
        content: to.content
    };
}
/**
 * Performs “inverted” diff between given documents: compares `from` and `to` documents
 * and returns `from` document with patched XML
 */
function diffInverted(from, to, options) {
    const diffs = getDiff(from, to, options);
    const state = createState(from.tokens);
    const fragmentOpt = {
        tags: options.preserveTags || []
    };
    let toOffset = 0;
    let fromOffset = 0;
    diffs.forEach(d => {
        let value = d[1];
        if (d[0] === diffMatchPatch.DIFF_DELETE && value) {
            if (suppressWhitespace(value, toOffset, state)) {
                fromOffset += 1;
                value = value.slice(1);
            }
            // if (!shouldSkipDel(to.content, value, fromOffset, options)) {
            //     // Unlike in “natural” diff, in ”inverted” deleted fragment if a part
            //     // of destination document
            // }
            moveSlice(from, fromOffset, fromOffset + value.length, 'del', state);
            fromOffset += value.length;
        }
        else if (d[0] === diffMatchPatch.DIFF_INSERT) {
            // Inserted fragment: should insert open and close tags at proper
            // positions and maintain valid XML nesting
            if (suppressWhitespace(value, toOffset, state)) {
                toOffset += 1;
                value = value.slice(1);
            }
            // if (shouldSkipIns(value, toOffset, state, options)) {
            //     state.push(whitespace(toOffset, value));
            // } else {
            // }
            moveTokensUntilPos(state, fromOffset);
            const chunk = fragment(to, toOffset, toOffset + value.length, fragmentOpt);
            state.push(chunk.toDiffToken('ins', value, fromOffset));
            toOffset += value.length;
        }
        else if (d[0] === diffMatchPatch.DIFF_EQUAL) {
            // Unmodified content
            toOffset += value.length;
            fromOffset += value.length;
            // Move all tokens of destination document to output result
            while (state.hasNext()) {
                const first = state.current();
                // if (first.location > offset || (first.location === offset && first.type === ElementType.Open)) {
                //     break;
                // }
                if (first.location > fromOffset) {
                    break;
                }
                if (first.location === fromOffset && first.type === 1 /* Open */) {
                    // Handle edge case. In the following examples:
                    // – aa <div>bb cc</div>
                    // – aa bb <div>cc</div>
                    // ...removing `bb ` results DELETE token with the same _text_
                    // location, yet in first case it should be outside `<div>` and
                    // in second case – inside `<div>`.
                    // In case if token touches the edge of open tag, we should detect
                    // if this token is inside or outside the same tag in `from`
                    // document
                    const { stack } = getElementStack(to.tokens, toOffset);
                    let found = false;
                    while (stack.length) {
                        if (stack.pop().name === first.name) {
                            found = true;
                            break;
                        }
                    }
                    if (!found) {
                        break;
                    }
                }
                state.push(first);
                state.ptr++;
            }
        }
    });
    return {
        tokens: state.output.concat(state.input.slice(state.ptr)),
        content: from.content
    };
}
/**
 * Moves sliced fragment from `model` document to `output`, starting at `tokenPos`
 * token location
 */
function moveSlice(model, from, to, tagName, state) {
    const chunk = slice(model, from, to, state.ptr);
    // Move tokens preceding sliced fragment to output
    while (state.ptr < chunk.range[0]) {
        state.push(model.tokens[state.ptr++]);
    }
    state.ptr = chunk.range[1];
    for (const t of chunk.toTokens(tagName)) {
        state.push(t);
    }
}
function moveTokensUntilPos(state, textPos) {
    while (state.hasNext()) {
        const t = state.current();
        if (t.location === textPos && (isType(t, 1 /* Open */) || (isType(t, 100 /* Space */) && !t.offset))) {
            state.push(t);
            state.ptr++;
        }
        else {
            break;
        }
    }
}
/**
 * Check if given INSERT patch should be omitted to reduce noise
 */
function shouldSkipIns(value, pos, state, options) {
    if (options.compact && isWhitespace(value)) {
        // Inserted whitespace.
        // It can be either significant (`ab` -> `a b`) or insignificant,
        // if this whitespace is empty or is right after non-inline element
        const prev = last(state.output);
        if (prev && isTagToken(prev) && prev.location === pos && !isInlineElement(prev, options)) {
            return true;
        }
    }
    return false;
}
/**
 * Check if given DELETE patch should be omitted
 */
function shouldSkipDel(content, value, pos, options) {
    if (options.compact && isWhitespace(value)) {
        if (isWhitespace(content.charAt(pos - 1)) || isWhitespace(content.charAt(pos))) {
            // There’s whitespace before or after deleted whitespace token
            return true;
        }
        // if (!isInlineElement(input[inputPos - 1], options) && !isInlineElement(input[inputPos], options)) {
        //     // Whitespace between non-inline elements
        //     return true;
        // }
    }
    return false;
}
/**
 * Check if given token is an inline-level HTML element
 */
function isInlineElement(token, options) {
    if (token) {
        return token.type === 101 /* Diff */
            || options.inlineElements.includes(token.name);
    }
    return false;
}
/**
 * Handle edge case when patch intersects with suppressed formatting whitespace,
 * used as word delimiter for adjacent blocks.
 * For example, `<div>a</div><div>b</div>` is represented as `a b` in plain text,
 * space between `a` and `b` is a suppressed whitespace and must be removed from
 * patch
 */
function suppressWhitespace(value, pos, state) {
    const lastToken = last(state.output);
    return lastToken
        ? isType(lastToken, 100 /* Space */)
            && !!lastToken.offset
            && lastToken.location === pos
            && value[0] === ' '
        : false;
}
function getDiff(from, to, options) {
    const dmp = new diffMatchPatch.diff_match_patch();
    if (options.dmp) {
        Object.assign(dmp, options.dmp);
    }
    let diffs = dmp.diff_main(from.content, to.content);
    dmp.diff_cleanupSemantic(diffs);
    if (options.replaceThreshold && getChangeThreshold(diffs) > options.replaceThreshold) {
        // Text is too different, mark it as replaced
        return [
            [diffMatchPatch.DIFF_DELETE, from.content],
            [diffMatchPatch.DIFF_INSERT, to.content],
        ];
    }
    if (options.wordPatches) {
        diffs = updateWordBounds(diffs);
    }
    return diffs;
}
function whitespace(location, value) {
    return {
        name: '#whitespace',
        type: 100 /* Space */,
        value,
        location,
    };
}
function createState(input) {
    return {
        ptr: 0,
        input,
        output: [],
        push(token) {
            this.output.push(token);
        },
        current() {
            return input[this.ptr];
        },
        hasNext() {
            return this.ptr < input.length;
        }
    };
}
function getChangeThreshold(diffs) {
    let i = 0;
    let changed = 0;
    let unchanged = 0;
    while (i < diffs.length) {
        const chunk = diffs[i++];
        let len = chunk[1].length;
        if (chunk[0] === diffMatchPatch.DIFF_EQUAL) {
            unchanged += len;
        }
        else {
            const next = diffs[i];
            if (chunk[0] === diffMatchPatch.DIFF_DELETE && next && next[0] === diffMatchPatch.DIFF_INSERT) {
                // Mark consecutive updates as single edit
                len = (len + next[1].length) / 2;
                i++;
            }
            changed += len;
        }
    }
    return changed / unchanged;
}

const reSpace = /[\s\r\n]+/g;
/** Normalized space character, used in text */
const textSpace = ' ';
/**
 * Parses given XML source into internal model for diff
 */
function parse(text, options = createOptions()) {
    const { consume, finalize } = createConsumer(text, options);
    htmlMatcher.scan(text, consume, options);
    return finalize(text.length);
}
/**
 * Creates XML markup consumer for given text. Returns object with `consume` function
 * that should receive markup tokens and `finalize` function to produce final
 * document model
 * @param text Original source code being parsed
 */
function createConsumer(text, options = createOptions()) {
    const baseStart = options.baseStart || 0;
    const state = {
        text,
        tokens: [],
        options,
        content: '',
        hasContent: false,
        prev: baseStart,
        offset: 0
    };
    return {
        state,
        consume(name, type, start, end) {
            if (state.options.normalizeSpace) {
                normalizeWhitespace(state.prev, start, state);
            }
            else {
                state.content += text.slice(state.prev, start);
            }
            const token = {
                name, type,
                value: text.substring(start, end),
                location: start - baseStart - state.offset,
                order: start
            };
            state.tokens.push(token);
            state.prev = end;
            state.offset += token.value.length;
            if (shouldAddSpaceDelimiter(name, type, state)) {
                // Closing block-level element. In case if `wordPatches` option enabled,
                // we should add whitespace to prevent word collapsing.
                // Add empty whitespace token into document to suppress whitespace
                // added to document content
                state.tokens.push({
                    name: '#whitespace',
                    type: 100 /* Space */,
                    value: '',
                    location: token.location,
                    offset: 1,
                    order: token.location
                });
                state.content += ' ';
                state.hasContent = false;
                state.offset--;
            }
        },
        finalize(baseEnd = text.length) {
            if (state.options.normalizeSpace) {
                normalizeWhitespace(state.prev, baseEnd, state);
                fixTrailingSpace(state);
            }
            else {
                state.content += text.slice(state.prev, baseEnd);
            }
            // Remove redundant whitespace tokens
            state.tokens = state.tokens.filter(token => {
                if (token.type === 100 /* Space */ && token.value === textSpace && token.offset === 1) {
                    return false;
                }
                return true;
            });
            return {
                tokens: state.tokens,
                content: state.content
            };
        }
    };
}
/**
 * Normalizes whitespace characters: replaces all whitespace occurrences
 * (including adjacent) with a single space and returns tokens with original space
 * characters
 */
function normalizeWhitespace(from, to, state) {
    let m;
    let prev = 0;
    let hasSpace = false;
    const fragment = state.text.substring(from, to);
    while (m = reSpace.exec(fragment)) {
        hasSpace = true;
        state.content += fragment.slice(prev, m.index);
        const token = {
            name: '#whitespace',
            type: 100 /* Space */,
            value: m[0],
            location: from + m.index - state.offset - (state.options.baseStart || 0),
            order: from + m.index
        };
        if (state.hasContent || m.index !== 0) {
            // Significant (not trailing) whitespace, replace it with single space
            state.content += textSpace;
            token.offset = 1;
            state.offset--;
        }
        prev = m.index + m[0].length;
        appendWhiteSpace(state, token);
        state.offset += token.value.length;
        state.hasContent = prev < fragment.length;
    }
    if (fragment && !hasSpace) {
        state.hasContent = true;
    }
    state.content += fragment.slice(prev);
}
/**
 * Fixes trailing space in accumulated content
 */
function fixTrailingSpace(state) {
    const lastCharIx = state.content.length - 1;
    const lastChar = state.content[lastCharIx];
    const trailing = lastChar === ' ';
    if (trailing) {
        // Offset trailing tokens with new space
        for (let i = state.tokens.length - 1; i >= 0; i--) {
            const token = state.tokens[i];
            if (token.location < lastCharIx) {
                break;
            }
            if (token.type === 100 /* Space */) {
                token.offset = 0;
                break;
            }
            else {
                token.location--;
            }
        }
        state.content = state.content.slice(0, -1);
        state.hasContent = state.content.length > 0;
    }
}
/**
 * Check if extra space delimiter should be added for the given XML token.
 * Used to determine if a space character should be added after closing tag in order
 * to prevent word collapsing in diff
 */
function shouldAddSpaceDelimiter(name, type, state) {
    return Boolean(state.options.wordPatches
        && (type === 2 /* Close */ || type === 3 /* SelfClose */)
        && !state.options.inlineElements.includes(name)
        && state.content
        && state.content.slice(-1) !== ' ');
}
/**
 * Adds given token to state
 */
function appendWhiteSpace(state, token) {
    const prev = last$1(state.tokens);
    if (prev && prev.type === 100 /* Space */ && !prev.value && prev.offset && prev.location + prev.offset === token.location) {
        // Merge whitespace tokens: previous one is a suppressing tag delimiter,
        // next one looks like formatting whitespace
        prev.value = token.value;
    }
    else {
        state.tokens.push(token);
    }
}
function last$1(arr) {
    return arr.length ? arr[arr.length - 1] : void 0;
}

/**
 * Calculates diff between two XML documents and returns `to` document content
 * with diff patches applied to it as `<ins>`/`<del>` tags
 */
function diffDocuments(from, to, options) {
    const opt = createOptions(options);
    const fromDoc = parse(from, opt);
    const toDoc = parse(to, opt);
    const diffDoc = diff(fromDoc, toDoc, opt);
    return stringify(diffDoc);
}
/**
 * Returns restored source from given parsed model
 */
function stringify(model) {
    let offset = 0;
    return model.tokens.map(token => {
        const result = model.content.slice(offset, token.location) + token.value;
        offset = token.location + (token.offset || 0);
        return result;
    }).join('') + model.content.slice(offset);
}

exports.createConsumer = createConsumer;
exports.createOptions = createOptions;
exports.default = diffDocuments;
exports.diff = diff;
exports.diffToken = diffToken;
exports.fragment = fragment;
exports.isDiffToken = isDiffToken;
exports.parse = parse;
exports.slice = slice;
exports.stringify = stringify;
//# sourceMappingURL=xml-diff.cjs.js.map
